/*
 *  Power BI Visualizations
 *
 *  Copyright (c) Microsoft Corporation
 *  All rights reserved.
 *  MIT License
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the ""Software""), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
var powerbi;
(function (powerbi) {
    var extensibility;
    (function (extensibility) {
        var utils;
        (function (utils) {
            var dataview;
            (function (dataview) {
                // TODO: refactor & focus DataViewTransform into a service with well-defined dependencies.
                var DataViewTransform;
                (function (DataViewTransform) {
                    // TODO: refactor this, setGrouped, and groupValues to a test helper to stop using it in the product
                    function createValueColumns(values, valueIdentityFields, source) {
                        if (values === void 0) { values = []; }
                        var result = values;
                        setGrouped(result);
                        if (valueIdentityFields) {
                            result.identityFields = valueIdentityFields;
                        }
                        if (source) {
                            result.source = source;
                        }
                        return result;
                    }
                    DataViewTransform.createValueColumns = createValueColumns;
                    function setGrouped(values, groupedResult) {
                        values.grouped = groupedResult
                            ? function () { return groupedResult; }
                            : function () { return groupValues(values); };
                    }
                    DataViewTransform.setGrouped = setGrouped;
                    /** Group together the values with a common identity. */
                    function groupValues(values) {
                        var groups = [], currentGroup;
                        for (var i = 0, len = values.length; i < len; i++) {
                            var value = values[i];
                            if (!currentGroup || currentGroup.identity !== value.identity) {
                                currentGroup = {
                                    values: []
                                };
                                if (value.identity) {
                                    currentGroup.identity = value.identity;
                                    var source = value.source;
                                    // allow null, which will be formatted as (Blank).
                                    if (source.groupName !== undefined) {
                                        currentGroup.name = source.groupName;
                                    }
                                    else if (source.displayName) {
                                        currentGroup.name = source.displayName;
                                    }
                                }
                                groups.push(currentGroup);
                            }
                            currentGroup.values.push(value);
                        }
                        return groups;
                    }
                    DataViewTransform.groupValues = groupValues;
                })(DataViewTransform = dataview.DataViewTransform || (dataview.DataViewTransform = {}));
            })(dataview = utils.dataview || (utils.dataview = {}));
        })(utils = extensibility.utils || (extensibility.utils = {}));
    })(extensibility = powerbi.extensibility || (powerbi.extensibility = {}));
})(powerbi || (powerbi = {}));
/*
 *  Power BI Visualizations
 *
 *  Copyright (c) Microsoft Corporation
 *  All rights reserved.
 *  MIT License
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the ""Software""), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
var powerbi;
(function (powerbi) {
    var extensibility;
    (function (extensibility) {
        var utils;
        (function (utils) {
            var dataview;
            (function (dataview) {
                var DataRoleHelper;
                (function (DataRoleHelper) {
                    function getMeasureIndexOfRole(grouped, roleName) {
                        if (!_.isEmpty(grouped)) {
                            var firstGroup = grouped[0];
                            if (firstGroup.values && firstGroup.values.length > 0) {
                                for (var i = 0, len = firstGroup.values.length; i < len; ++i) {
                                    var value = firstGroup.values[i];
                                    if (value && value.source) {
                                        if (hasRole(value.source, roleName)) {
                                            return i;
                                        }
                                    }
                                }
                            }
                        }
                        return -1;
                    }
                    DataRoleHelper.getMeasureIndexOfRole = getMeasureIndexOfRole;
                    function getCategoryIndexOfRole(categories, roleName) {
                        if (!_.isEmpty(categories)) {
                            for (var i = 0, ilen = categories.length; i < ilen; i++) {
                                if (hasRole(categories[i].source, roleName)) {
                                    return i;
                                }
                            }
                        }
                        return -1;
                    }
                    DataRoleHelper.getCategoryIndexOfRole = getCategoryIndexOfRole;
                    function hasRole(column, name) {
                        var roles = column.roles;
                        return roles && roles[name];
                    }
                    DataRoleHelper.hasRole = hasRole;
                    function hasRoleInDataView(dataView, name) {
                        return dataView != null
                            && dataView.metadata != null
                            && dataView.metadata.columns
                            && _.some(dataView.metadata.columns, function (c) { return c.roles && c.roles[name] !== undefined; }); // any is an alias of some
                    }
                    DataRoleHelper.hasRoleInDataView = hasRoleInDataView;
                    function hasRoleInValueColumn(valueColumn, name) {
                        return valueColumn
                            && valueColumn.source
                            && valueColumn.source.roles
                            && (valueColumn.source.roles[name] === true);
                    }
                    DataRoleHelper.hasRoleInValueColumn = hasRoleInValueColumn;
                })(DataRoleHelper = dataview.DataRoleHelper || (dataview.DataRoleHelper = {}));
            })(dataview = utils.dataview || (utils.dataview = {}));
        })(utils = extensibility.utils || (extensibility.utils = {}));
    })(extensibility = powerbi.extensibility || (powerbi.extensibility = {}));
})(powerbi || (powerbi = {}));
/*
 *  Power BI Visualizations
 *
 *  Copyright (c) Microsoft Corporation
 *  All rights reserved.
 *  MIT License
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the ""Software""), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
var powerbi;
(function (powerbi) {
    var extensibility;
    (function (extensibility) {
        var utils;
        (function (utils) {
            var dataview;
            (function (dataview) {
                var DataViewObject;
                (function (DataViewObject) {
                    function getValue(object, propertyName, defaultValue) {
                        if (!object) {
                            return defaultValue;
                        }
                        var propertyValue = object[propertyName];
                        if (propertyValue === undefined) {
                            return defaultValue;
                        }
                        return propertyValue;
                    }
                    DataViewObject.getValue = getValue;
                    /** Gets the solid color from a fill property using only a propertyName */
                    function getFillColorByPropertyName(object, propertyName, defaultColor) {
                        var value = getValue(object, propertyName);
                        if (!value || !value.solid) {
                            return defaultColor;
                        }
                        return value.solid.color;
                    }
                    DataViewObject.getFillColorByPropertyName = getFillColorByPropertyName;
                })(DataViewObject = dataview.DataViewObject || (dataview.DataViewObject = {}));
            })(dataview = utils.dataview || (utils.dataview = {}));
        })(utils = extensibility.utils || (extensibility.utils = {}));
    })(extensibility = powerbi.extensibility || (powerbi.extensibility = {}));
})(powerbi || (powerbi = {}));
/*
 *  Power BI Visualizations
 *
 *  Copyright (c) Microsoft Corporation
 *  All rights reserved.
 *  MIT License
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the ""Software""), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
var powerbi;
(function (powerbi) {
    var extensibility;
    (function (extensibility) {
        var utils;
        (function (utils) {
            var dataview;
            (function (dataview) {
                var DataViewObjects;
                (function (DataViewObjects) {
                    /** Gets the value of the given object/property pair. */
                    function getValue(objects, propertyId, defaultValue) {
                        if (!objects) {
                            return defaultValue;
                        }
                        return dataview.DataViewObject.getValue(objects[propertyId.objectName], propertyId.propertyName, defaultValue);
                    }
                    DataViewObjects.getValue = getValue;
                    /** Gets an object from objects. */
                    function getObject(objects, objectName, defaultValue) {
                        if (objects && objects[objectName]) {
                            return objects[objectName];
                        }
                        return defaultValue;
                    }
                    DataViewObjects.getObject = getObject;
                    /** Gets the solid color from a fill property. */
                    function getFillColor(objects, propertyId, defaultColor) {
                        var value = getValue(objects, propertyId);
                        if (!value || !value.solid) {
                            return defaultColor;
                        }
                        return value.solid.color;
                    }
                    DataViewObjects.getFillColor = getFillColor;
                    function getCommonValue(objects, propertyId, defaultValue) {
                        var value = getValue(objects, propertyId, defaultValue);
                        if (value && value.solid) {
                            return value.solid.color;
                        }
                        if (value === undefined
                            || value === null
                            || (typeof value === "object" && !value.solid)) {
                            return defaultValue;
                        }
                        return value;
                    }
                    DataViewObjects.getCommonValue = getCommonValue;
                })(DataViewObjects = dataview.DataViewObjects || (dataview.DataViewObjects = {}));
            })(dataview = utils.dataview || (utils.dataview = {}));
        })(utils = extensibility.utils || (extensibility.utils = {}));
    })(extensibility = powerbi.extensibility || (powerbi.extensibility = {}));
})(powerbi || (powerbi = {}));
/*
 *  Power BI Visualizations
 *
 *  Copyright (c) Microsoft Corporation
 *  All rights reserved.
 *  MIT License
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the ""Software""), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
var powerbi;
(function (powerbi) {
    var extensibility;
    (function (extensibility) {
        var utils;
        (function (utils) {
            var dataview;
            (function (dataview) {
                // powerbi.extensibility.utils.dataview
                var DataRoleHelper = powerbi.extensibility.utils.dataview.DataRoleHelper;
                var converterHelper;
                (function (converterHelper) {
                    function categoryIsAlsoSeriesRole(dataView, seriesRoleName, categoryRoleName) {
                        if (dataView.categories && dataView.categories.length > 0) {
                            // Need to pivot data if our category soure is a series role
                            var category = dataView.categories[0];
                            return category.source &&
                                DataRoleHelper.hasRole(category.source, seriesRoleName) &&
                                DataRoleHelper.hasRole(category.source, categoryRoleName);
                        }
                        return false;
                    }
                    converterHelper.categoryIsAlsoSeriesRole = categoryIsAlsoSeriesRole;
                    function getSeriesName(source) {
                        return (source.groupName !== undefined)
                            ? source.groupName
                            : source.queryName;
                    }
                    converterHelper.getSeriesName = getSeriesName;
                    function isImageUrlColumn(column) {
                        var misc = getMiscellaneousTypeDescriptor(column);
                        return misc != null && misc.imageUrl === true;
                    }
                    converterHelper.isImageUrlColumn = isImageUrlColumn;
                    function isWebUrlColumn(column) {
                        var misc = getMiscellaneousTypeDescriptor(column);
                        return misc != null && misc.webUrl === true;
                    }
                    converterHelper.isWebUrlColumn = isWebUrlColumn;
                    function getMiscellaneousTypeDescriptor(column) {
                        return column
                            && column.type
                            && column.type.misc;
                    }
                    converterHelper.getMiscellaneousTypeDescriptor = getMiscellaneousTypeDescriptor;
                    function hasImageUrlColumn(dataView) {
                        if (!dataView || !dataView.metadata || _.isEmpty(dataView.metadata.columns))
                            return false;
                        return _.some(dataView.metadata.columns, function (column) { return isImageUrlColumn(column) === true; });
                    }
                    converterHelper.hasImageUrlColumn = hasImageUrlColumn;
                })(converterHelper = dataview.converterHelper || (dataview.converterHelper = {}));
            })(dataview = utils.dataview || (utils.dataview = {}));
        })(utils = extensibility.utils || (extensibility.utils = {}));
    })(extensibility = powerbi.extensibility || (powerbi.extensibility = {}));
})(powerbi || (powerbi = {}));
/*
 *  Power BI Visualizations
 *
 *  Copyright (c) Microsoft Corporation
 *  All rights reserved.
 *  MIT License
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the ""Software""), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
var powerbi;
(function (powerbi) {
    var extensibility;
    (function (extensibility) {
        var utils;
        (function (utils) {
            var dataview;
            (function (dataview) {
                var DataViewObjectsParser = (function () {
                    function DataViewObjectsParser() {
                    }
                    DataViewObjectsParser.getDefault = function () {
                        return new this();
                    };
                    DataViewObjectsParser.createPropertyIdentifier = function (objectName, propertyName) {
                        return {
                            objectName: objectName,
                            propertyName: propertyName
                        };
                    };
                    DataViewObjectsParser.parse = function (dataView) {
                        var dataViewObjectParser = this.getDefault(), properties;
                        if (!dataView || !dataView.metadata || !dataView.metadata.objects) {
                            return dataViewObjectParser;
                        }
                        properties = dataViewObjectParser.getProperties();
                        for (var objectName in properties) {
                            for (var propertyName in properties[objectName]) {
                                var defaultValue = dataViewObjectParser[objectName][propertyName];
                                dataViewObjectParser[objectName][propertyName] = dataview.DataViewObjects.getCommonValue(dataView.metadata.objects, properties[objectName][propertyName], defaultValue);
                            }
                        }
                        return dataViewObjectParser;
                    };
                    DataViewObjectsParser.isPropertyEnumerable = function (propertyName) {
                        return !DataViewObjectsParser.InnumerablePropertyPrefix.test(propertyName);
                    };
                    DataViewObjectsParser.enumerateObjectInstances = function (dataViewObjectParser, options) {
                        var dataViewProperties = dataViewObjectParser && dataViewObjectParser[options.objectName];
                        if (!dataViewProperties) {
                            return [];
                        }
                        var instance = {
                            objectName: options.objectName,
                            selector: null,
                            properties: {}
                        };
                        for (var key in dataViewProperties) {
                            if (_.has(dataViewProperties, key)) {
                                instance.properties[key] = dataViewProperties[key];
                            }
                        }
                        return {
                            instances: [instance]
                        };
                    };
                    DataViewObjectsParser.prototype.getProperties = function () {
                        var _this = this;
                        var properties = {}, objectNames = Object.keys(this);
                        objectNames.forEach(function (objectName) {
                            if (DataViewObjectsParser.isPropertyEnumerable(objectName)) {
                                var propertyNames = Object.keys(_this[objectName]);
                                properties[objectName] = {};
                                propertyNames.forEach(function (propertyName) {
                                    if (DataViewObjectsParser.isPropertyEnumerable(objectName)) {
                                        properties[objectName][propertyName] =
                                            DataViewObjectsParser.createPropertyIdentifier(objectName, propertyName);
                                    }
                                });
                            }
                        });
                        return properties;
                    };
                    DataViewObjectsParser.InnumerablePropertyPrefix = /^_/;
                    return DataViewObjectsParser;
                }());
                dataview.DataViewObjectsParser = DataViewObjectsParser;
            })(dataview = utils.dataview || (utils.dataview = {}));
        })(utils = extensibility.utils || (extensibility.utils = {}));
    })(extensibility = powerbi.extensibility || (powerbi.extensibility = {}));
})(powerbi || (powerbi = {}));

var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
/**
 *  Power BI Visualizations
 *
 *  Copyright (c) Microsoft Corporation
 *  All rights reserved.
 *  MIT License
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the ""Software""), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
var powerbi;
(function (powerbi) {
    var visuals;
    (function (visuals) {
        var PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0;
        (function (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0) {
            var dualKpi;
            (function (dualKpi) {
                var DualKpiAxisSettings = /** @class */ (function () {
                    function DualKpiAxisSettings() {
                        this.topChartAxisMin = null;
                        this.topChartAxisMax = null;
                        this.bottomChartAxisMin = null;
                        this.bottomChartAxisMax = null;
                        this.topChartZeroLine = false;
                        this.bottomChartZeroLine = false;
                    }
                    return DualKpiAxisSettings;
                }());
                dualKpi.DualKpiAxisSettings = DualKpiAxisSettings;
            })(dualKpi = PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.dualKpi || (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.dualKpi = {}));
        })(PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 || (visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = {}));
    })(visuals = powerbi.visuals || (powerbi.visuals = {}));
})(powerbi || (powerbi = {}));
/**
 *  Power BI Visualizations
 *
 *  Copyright (c) Microsoft Corporation
 *  All rights reserved.
 *  MIT License
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the ""Software""), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
var powerbi;
(function (powerbi) {
    var visuals;
    (function (visuals) {
        var PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0;
        (function (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0) {
            var dualKpi;
            (function (dualKpi) {
                var DualKpiChartSettings = /** @class */ (function () {
                    function DualKpiChartSettings() {
                        this.topChartType = "area";
                        this.bottomChartType = "area";
                    }
                    return DualKpiChartSettings;
                }());
                dualKpi.DualKpiChartSettings = DualKpiChartSettings;
            })(dualKpi = PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.dualKpi || (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.dualKpi = {}));
        })(PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 || (visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = {}));
    })(visuals = powerbi.visuals || (powerbi.visuals = {}));
})(powerbi || (powerbi = {}));
/**
 *  Power BI Visualizations
 *
 *  Copyright (c) Microsoft Corporation
 *  All rights reserved.
 *  MIT License
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the ""Software""), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
var powerbi;
(function (powerbi) {
    var visuals;
    (function (visuals) {
        var PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0;
        (function (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0) {
            var dualKpi;
            (function (dualKpi) {
                var DualKpiColorsSettings = /** @class */ (function () {
                    function DualKpiColorsSettings() {
                        this.dataColor = "#01b8aa";
                        this.textColor = "#212121";
                        this.opacity = 30;
                    }
                    return DualKpiColorsSettings;
                }());
                dualKpi.DualKpiColorsSettings = DualKpiColorsSettings;
            })(dualKpi = PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.dualKpi || (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.dualKpi = {}));
        })(PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 || (visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = {}));
    })(visuals = powerbi.visuals || (powerbi.visuals = {}));
})(powerbi || (powerbi = {}));
/**
 *  Power BI Visualizations
 *
 *  Copyright (c) Microsoft Corporation
 *  All rights reserved.
 *  MIT License
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the ""Software""), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
var powerbi;
(function (powerbi) {
    var visuals;
    (function (visuals) {
        var PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0;
        (function (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0) {
            var dualKpi;
            (function (dualKpi) {
                var DualKpiColorsBottomSettings = /** @class */ (function () {
                    function DualKpiColorsBottomSettings() {
                        this.matchTopChartOptions = true;
                        this.dataColor = "#01b8aa";
                        this.textColor = "#212121";
                        this.opacity = 30;
                    }
                    return DualKpiColorsBottomSettings;
                }());
                dualKpi.DualKpiColorsBottomSettings = DualKpiColorsBottomSettings;
            })(dualKpi = PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.dualKpi || (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.dualKpi = {}));
        })(PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 || (visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = {}));
    })(visuals = powerbi.visuals || (powerbi.visuals = {}));
})(powerbi || (powerbi = {}));
/**
 *  Power BI Visualizations
 *
 *  Copyright (c) Microsoft Corporation
 *  All rights reserved.
 *  MIT License
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the ""Software""), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
var powerbi;
(function (powerbi) {
    var visuals;
    (function (visuals) {
        var PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0;
        (function (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0) {
            var dualKpi;
            (function (dualKpi) {
                var DualKpiFormattingSettings = /** @class */ (function () {
                    function DualKpiFormattingSettings() {
                        this.fontSizeAutoFormatting = true;
                        this.fontSize = 12;
                        this.isBold = false;
                        this.isItalic = false;
                        this.fontFamily = "helvetica, arial, sans-serif";
                    }
                    return DualKpiFormattingSettings;
                }());
                dualKpi.DualKpiFormattingSettings = DualKpiFormattingSettings;
            })(dualKpi = PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.dualKpi || (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.dualKpi = {}));
        })(PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 || (visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = {}));
    })(visuals = powerbi.visuals || (powerbi.visuals = {}));
})(powerbi || (powerbi = {}));
/**
 *  Power BI Visualizations
 *
 *  Copyright (c) Microsoft Corporation
 *  All rights reserved.
 *  MIT License
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the ""Software""), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
var powerbi;
(function (powerbi) {
    var visuals;
    (function (visuals) {
        var PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0;
        (function (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0) {
            var dualKpi;
            (function (dualKpi) {
                var DualKpiPropertiesSettings = /** @class */ (function () {
                    function DualKpiPropertiesSettings() {
                        this.topChartShow = true;
                        this.bottomChartShow = true;
                        this.titleText = "Title";
                        this.abbreviateValues = false;
                        this.abbreviateHoverValues = false;
                        this.shortKpiTooltip = false;
                        this.hoverDataPercentType = PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.PercentType.lastDate;
                        this.topChartToolTipText = "";
                        this.bottomChartToolTipText = "";
                        this.warningTooltipText = "Warning message";
                        this.showStaleDataWarning = true;
                        this.staleDataTooltipText = "";
                        this.staleDataThreshold = 2;
                        this.topPercentCalcDate = null;
                        this.bottomPercentCalcDate = null;
                    }
                    return DualKpiPropertiesSettings;
                }());
                dualKpi.DualKpiPropertiesSettings = DualKpiPropertiesSettings;
            })(dualKpi = PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.dualKpi || (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.dualKpi = {}));
        })(PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 || (visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = {}));
    })(visuals = powerbi.visuals || (powerbi.visuals = {}));
})(powerbi || (powerbi = {}));
/**
 *  Power BI Visualizations
 *
 *  Copyright (c) Microsoft Corporation
 *  All rights reserved.
 *  MIT License
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the ""Software""), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
var powerbi;
(function (powerbi) {
    var visuals;
    (function (visuals) {
        var PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0;
        (function (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0) {
            var dualKpi;
            (function (dualKpi) {
                var DualKpiTitleFormattingSettings = /** @class */ (function (_super) {
                    __extends(DualKpiTitleFormattingSettings, _super);
                    function DualKpiTitleFormattingSettings() {
                        var _this = _super !== null && _super.apply(this, arguments) || this;
                        _this.fontSize = 10;
                        return _this;
                    }
                    return DualKpiTitleFormattingSettings;
                }(dualKpi.DualKpiFormattingSettings));
                dualKpi.DualKpiTitleFormattingSettings = DualKpiTitleFormattingSettings;
            })(dualKpi = PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.dualKpi || (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.dualKpi = {}));
        })(PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 || (visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = {}));
    })(visuals = powerbi.visuals || (powerbi.visuals = {}));
})(powerbi || (powerbi = {}));
/**
 *  Power BI Visualizations
 *
 *  Copyright (c) Microsoft Corporation
 *  All rights reserved.
 *  MIT License
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the ""Software""), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
var powerbi;
(function (powerbi) {
    var visuals;
    (function (visuals) {
        var PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0;
        (function (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0) {
            var dualKpi;
            (function (dualKpi) {
                var DualKpiValueFormattingSettings = /** @class */ (function (_super) {
                    __extends(DualKpiValueFormattingSettings, _super);
                    function DualKpiValueFormattingSettings() {
                        var _this = _super !== null && _super.apply(this, arguments) || this;
                        _this.fontSize = 16;
                        _this.displayUnits = 1;
                        _this.precision = 0;
                        return _this;
                    }
                    return DualKpiValueFormattingSettings;
                }(dualKpi.DualKpiFormattingSettings));
                dualKpi.DualKpiValueFormattingSettings = DualKpiValueFormattingSettings;
            })(dualKpi = PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.dualKpi || (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.dualKpi = {}));
        })(PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 || (visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = {}));
    })(visuals = powerbi.visuals || (powerbi.visuals = {}));
})(powerbi || (powerbi = {}));
/**
 *  Power BI Visualizations
 *
 *  Copyright (c) Microsoft Corporation
 *  All rights reserved.
 *  MIT License
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the ""Software""), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
var powerbi;
(function (powerbi) {
    var visuals;
    (function (visuals) {
        var PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0;
        (function (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0) {
            var dualKpi;
            (function (dualKpi) {
                var DualKpiValuesSettings = /** @class */ (function () {
                    function DualKpiValuesSettings() {
                        this.show = true;
                        this.showKpiValuesTop = true;
                        this.topChartDefaultKpiValue = "N/A";
                        this.showKpiValuesBottom = true;
                        this.bottomChartDefaultKpiValue = "N/A";
                    }
                    return DualKpiValuesSettings;
                }());
                dualKpi.DualKpiValuesSettings = DualKpiValuesSettings;
            })(dualKpi = PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.dualKpi || (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.dualKpi = {}));
        })(PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 || (visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = {}));
    })(visuals = powerbi.visuals || (powerbi.visuals = {}));
})(powerbi || (powerbi = {}));
/**
 *  Power BI Visualizations
 *
 *  Copyright (c) Microsoft Corporation
 *  All rights reserved.
 *  MIT License
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the ""Software""), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
var powerbi;
(function (powerbi) {
    var visuals;
    (function (visuals) {
        var PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0;
        (function (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0) {
            var dualKpi;
            (function (dualKpi) {
                // powerbi.extensibility.utils.dataview
                var DataViewObjectsParser = powerbi.extensibility.utils.dataview.DataViewObjectsParser;
                var DualKpiSettings = /** @class */ (function (_super) {
                    __extends(DualKpiSettings, _super);
                    function DualKpiSettings() {
                        var _this = _super !== null && _super.apply(this, arguments) || this;
                        _this.dualKpiProperties = new dualKpi.DualKpiPropertiesSettings();
                        _this.dualKpiValues = new dualKpi.DualKpiValuesSettings();
                        _this.dualKpiTitleFormatting = new dualKpi.DualKpiTitleFormattingSettings();
                        _this.dualKpiValueFormatting = new dualKpi.DualKpiValueFormattingSettings();
                        _this.dualKpiColors = new dualKpi.DualKpiColorsSettings();
                        _this.dualKpiColorsBottom = new dualKpi.DualKpiColorsBottomSettings();
                        _this.dualKpiAxis = new dualKpi.DualKpiAxisSettings();
                        _this.dualKpiChart = new dualKpi.DualKpiChartSettings();
                        return _this;
                    }
                    return DualKpiSettings;
                }(DataViewObjectsParser));
                dualKpi.DualKpiSettings = DualKpiSettings;
            })(dualKpi = PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.dualKpi || (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.dualKpi = {}));
        })(PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 || (visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = {}));
    })(visuals = powerbi.visuals || (powerbi.visuals = {}));
})(powerbi || (powerbi = {}));
/*
 *  Power BI Visualizations
 *
 *  Copyright (c) Microsoft Corporation
 *  All rights reserved.
 *  MIT License
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the ""Software""), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */

/* tslint:disable:max-file-line-count */
/* Do not copy this TSLint exclusion onto additional files. */
var powerbi;
(function (powerbi) {
    var visuals;
    (function (visuals) {
        var PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0;
        (function (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0) {
            var DualKpiSettings = PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.dualKpi.DualKpiSettings;
            var textMeasurementService = powerbi.TextMeasurementService;
            var valueFormatter = powerbi.visuals.valueFormatter;
            var DualKpiSize;
            (function (DualKpiSize) {
                DualKpiSize[DualKpiSize["supersmall"] = 0] = "supersmall";
                DualKpiSize[DualKpiSize["extrasmall"] = 1] = "extrasmall";
                DualKpiSize[DualKpiSize["small"] = 2] = "small";
                DualKpiSize[DualKpiSize["medium"] = 3] = "medium";
                DualKpiSize[DualKpiSize["large"] = 4] = "large";
            })(DualKpiSize = PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.DualKpiSize || (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.DualKpiSize = {}));
            var PercentType;
            (function (PercentType) {
                PercentType[PercentType["lastDate"] = "lastDate"] = "lastDate";
                PercentType[PercentType["firstDate"] = "firstDate"] = "firstDate";
                PercentType[PercentType["previousDate"] = "previousDate"] = "previousDate";
            })(PercentType = PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.PercentType || (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.PercentType = {}));
            var DualKpiChartPositionType;
            (function (DualKpiChartPositionType) {
                DualKpiChartPositionType[DualKpiChartPositionType["top"] = "top"] = "top";
                DualKpiChartPositionType[DualKpiChartPositionType["bottom"] = "bottom"] = "bottom";
            })(DualKpiChartPositionType = PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.DualKpiChartPositionType || (PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.DualKpiChartPositionType = {}));
            var DualKpiHoverDataPercentType;
            (function (DualKpiHoverDataPercentType) {
                DualKpiHoverDataPercentType.LAST_DATE = 'Last date';
                DualKpiHoverDataPercentType.FIRST_DATE = 'First date';
                DualKpiHoverDataPercentType.PREVIOUS_DATE = 'Previous date';
                DualKpiHoverDataPercentType.type = powerbi.createEnumType([
                    { value: PercentType.lastDate, displayName: DualKpiHoverDataPercentType.LAST_DATE },
                    { value: PercentType.firstDate, displayName: DualKpiHoverDataPercentType.FIRST_DATE },
                    { value: PercentType.previousDate, displayName: DualKpiHoverDataPercentType.PREVIOUS_DATE }
                ]);
            })(DualKpiHoverDataPercentType || (DualKpiHoverDataPercentType = {}));
            var DualKpiChartType;
            (function (DualKpiChartType) {
                DualKpiChartType.AREA = 'Area';
                DualKpiChartType.LINE = 'Line';
                DualKpiChartType.type = powerbi.createEnumType([
                    { value: "area", displayName: DualKpiChartType.AREA },
                    { value: "line", displayName: DualKpiChartType.LINE }
                ]);
            })(DualKpiChartType || (DualKpiChartType = {}));
            var DualKpi = /** @class */ (function () {
                function DualKpi() {
                    this.chartLeftMargin = 35;
                    this.touchEventsEnabled = false;
                    this.titleHeight = 0;
                }
                DualKpi.prototype.init = function (options) {
                    this.target = options.element;
                    this.size = DualKpiSize.small;
                    this.sizeCssClass = "small";
                    this.valueFormatter = d3.format(".3s");
                    this.axisNumberFormatter = d3.format(".2s");
                    this.commaNumberFormatter = d3.format(",");
                    this.timeFormatter = d3.time.format("%m/%d/%y");
                    this.dataBisector = d3.bisector(function (d) { return d.date; }).left;
                    this.dispatch = d3.dispatch("onDualKpiMouseMove", "onDualKpiMouseOut");
                    this.initContainer();
                    this.initMouseEvents();
                };
                DualKpi.prototype.initMouseEvents = function () {
                    var dispatch = this.dispatch;
                    var target = this.target;
                    var targetElement = d3.select(target[0]);
                    var onMouseMove = function (e) {
                        dispatch.onDualKpiMouseMove(d3.mouse(this));
                    };
                    targetElement.on("mousemove", onMouseMove);
                    targetElement.on("mouseout", onMouseMove);
                    targetElement.on("touchmove", onMouseMove);
                    targetElement.on("touchstart", onMouseMove);
                    var onMouseOut = function (e) {
                        dispatch.onDualKpiMouseOut();
                    };
                    targetElement.on("mouseleave", onMouseOut);
                    targetElement.on("touchleave", onMouseOut);
                };
                DualKpi.prototype.initContainer = function () {
                    var xmlns = "http://www.w3.org/2000/svg";
                    var svgElem = document.createElementNS(xmlns, "svg");
                    var svgRoot = this.svgRoot = d3.select(svgElem);
                    svgRoot
                        .attr("class", "dualKpi");
                    this.chartGroupTop = this.createChartGroup(svgRoot, DualKpiChartPositionType.top);
                    this.chartGroupBottom = this.createChartGroup(svgRoot, DualKpiChartPositionType.bottom);
                    this.bottomContainer = this.createBottomContainer(svgRoot);
                    this.target.append(svgElem);
                };
                DualKpi.prototype.createBottomContainer = function (svgRoot) {
                    var bottomContainer = svgRoot
                        .append("g")
                        .attr("class", "bottom-title-container")
                        .classed(DualKpi.INVISIBLE, true);
                    var chartTitleElement = bottomContainer
                        .append("text")
                        .classed("title", true);
                    var warningGroup = bottomContainer
                        .append("g")
                        .classed("warning-group", true);
                    var warningIcon = warningGroup
                        .append("path")
                        .classed("warning-icon", true);
                    var warningTitle = warningGroup
                        .append("title")
                        .classed("warning-title", true);
                    var infoGroup = bottomContainer
                        .append("g")
                        .classed("info-group", true);
                    var infoIcon = infoGroup
                        .append("path")
                        .classed("info-icon", true);
                    var infoTitle = infoIcon
                        .append("title")
                        .classed("info-title", true);
                    var dateRangeText = bottomContainer
                        .append("text")
                        .classed("date-range-text", true)
                        .attr("text-anchor", "end");
                    return {
                        bottomContainer: bottomContainer,
                        chartTitleElement: chartTitleElement,
                        warning: {
                            group: warningGroup,
                            icon: warningIcon,
                            title: warningTitle
                        },
                        info: {
                            group: infoGroup,
                            icon: infoIcon,
                            title: infoTitle
                        },
                        dateRangeText: dateRangeText
                    };
                };
                DualKpi.prototype.createChartGroup = function (svgRoot, type) {
                    var chartGroup = svgRoot
                        .append("g")
                        .attr("class", "chartGroup")
                        .classed(type === DualKpiChartPositionType.top ? "chartGroupTop" : "chartGroupBottom", true);
                    var chartArea = chartGroup
                        .append("path")
                        .attr("class", "area");
                    var yAxis = chartGroup
                        .append("g")
                        .attr("class", "axis");
                    var hoverLine = chartGroup
                        .append("line")
                        .attr("class", "hoverLine");
                    var hoverDataContainer = this.createHoverDataContainer(chartGroup);
                    var chartOverlay = this.createChartOverlay(chartGroup);
                    var zeroAxis = chartGroup
                        .append("path")
                        .attr("class", "zero-axis");
                    return {
                        group: chartGroup,
                        area: chartArea,
                        yAxis: yAxis,
                        hoverLine: hoverLine,
                        hoverDataContainer: hoverDataContainer,
                        chartOverlay: chartOverlay,
                        zeroAxis: zeroAxis
                    };
                };
                DualKpi.prototype.createChartOverlay = function (chartGroup) {
                    var chartOverlayTextGroup = chartGroup
                        .append("g")
                        .classed("group", true);
                    var title = chartOverlayTextGroup
                        .append("text")
                        .classed("data-title", true)
                        .attr({
                        "text-anchor": "middle"
                    });
                    var text = chartOverlayTextGroup
                        .append("text")
                        .classed("data-value", true)
                        .attr({
                        "text-anchor": "middle"
                    });
                    // this rect is always invisible, used for capture mouse and touch events
                    var chartOverlayRect = chartOverlayTextGroup
                        .append("rect")
                        .attr("style", "stroke: none; fill: #000;opacity:0;");
                    var rectTitle = chartOverlayTextGroup
                        .append("title");
                    return {
                        group: chartOverlayTextGroup,
                        title: title,
                        text: text,
                        rect: chartOverlayRect,
                        rectTitle: rectTitle
                    };
                };
                DualKpi.getChartSize = function (viewport) {
                    var height = viewport.height, width = viewport.width;
                    if (width < 215 || height < 110) {
                        return DualKpiSize.supersmall;
                    }
                    if (width < 245 || height < 130) {
                        return DualKpiSize.extrasmall;
                    }
                    if (width < 350 || height < 280) {
                        return DualKpiSize.small;
                    }
                    if (width < 450 || height < 450) {
                        return DualKpiSize.medium;
                    }
                    return DualKpiSize.large;
                };
                DualKpi.prototype.update = function (options) {
                    var dataView = this.dataView = options.dataViews && options.dataViews[0];
                    if (!dataView ||
                        !dataView.metadata ||
                        !dataView.metadata.columns) {
                        this.displayRootElement(false);
                        return;
                    }
                    var data = this.data = DualKpi.converter(this.dataView);
                    var availableHeight = options.viewport.height < 90 ? 90 : options.viewport.height, availableWidth = options.viewport.width < 200 ? 200 : options.viewport.width, chartWidth = availableWidth, chartSpaceBetween, chartTitleSpace, iconOffset;
                    var size = DualKpi.getChartSize({ height: availableHeight, width: availableWidth });
                    switch (size) {
                        case DualKpiSize.large:
                            this.size = DualKpiSize.large;
                            this.sizeCssClass = "large";
                            iconOffset = -1;
                            chartSpaceBetween = 25;
                            chartTitleSpace = 46;
                            break;
                        case DualKpiSize.medium:
                            this.size = DualKpiSize.medium;
                            this.sizeCssClass = "medium";
                            iconOffset = -8;
                            chartSpaceBetween = 20;
                            chartTitleSpace = 30;
                            break;
                        case DualKpiSize.small:
                            this.size = DualKpiSize.small;
                            this.sizeCssClass = "small";
                            iconOffset = -6;
                            chartSpaceBetween = 15;
                            chartTitleSpace = 22;
                            break;
                        case DualKpiSize.extrasmall:
                            this.size = DualKpiSize.extrasmall;
                            this.sizeCssClass = "extra-small";
                            iconOffset = -8;
                            chartSpaceBetween = 12;
                            chartTitleSpace = 20;
                            break;
                        case DualKpiSize.supersmall:
                            this.size = DualKpiSize.supersmall;
                            this.sizeCssClass = "super-small";
                            iconOffset = -8;
                            chartSpaceBetween = 10;
                            chartTitleSpace = 18;
                            break;
                        default:
                            break;
                    }
                    this.titleHeight = DualKpi.DefaultTitleSizes[this.sizeCssClass];
                    this.updateViewport({
                        width: availableWidth,
                        height: availableHeight
                    });
                    var chartHeight = (availableHeight - (chartSpaceBetween + chartTitleSpace)) / 2;
                    var topChartAxisConfig = { min: data.settings.dualKpiAxis.topChartAxisMin, max: data.settings.dualKpiAxis.topChartAxisMax };
                    var bottomChartAxisConfig = { min: data.settings.dualKpiAxis.bottomChartAxisMin, max: data.settings.dualKpiAxis.bottomChartAxisMax };
                    var topChartPercentChangeStartPoint = DualKpi.getPercentChangeStartPoint(data.topValues, data.topPercentCalcDate);
                    var bottomChartPercentChangeStartPoint = DualKpi.getPercentChangeStartPoint(data.bottomValues, data.bottomPercentCalcDate);
                    var wasDataSetRendered = data.topValues.length > 0 || data.bottomValues.length > 0;
                    this.displayRootElement(wasDataSetRendered);
                    if (data.settings.dualKpiProperties.topChartShow) {
                        // draw top chart
                        this.chartGroupTop.group.classed(DualKpi.INVISIBLE, false);
                        this.chartGroupTop.hoverDataContainer.container.classed(DualKpi.INVISIBLE, true);
                        if (data.topValues.length > 0) {
                            this.drawChart({
                                element: this.chartGroupTop,
                                abbreviateValue: data.settings.dualKpiProperties.abbreviateValues,
                                abbreviateHoverValue: data.settings.dualKpiProperties.abbreviateHoverValues,
                                hoverDataPercentType: data.settings.dualKpiProperties.hoverDataPercentType,
                                axisConfig: topChartAxisConfig,
                                chartData: data.topValues,
                                chartTitle: data.topChartName,
                                chartType: data.settings.dualKpiChart.topChartType,
                                height: data.settings.dualKpiProperties.bottomChartShow && data.settings.dualKpiProperties.topChartShow ? chartHeight : chartHeight * 2 + chartSpaceBetween,
                                percentChangeStartPoint: topChartPercentChangeStartPoint,
                                showZeroLine: data.settings.dualKpiAxis.topChartZeroLine,
                                tooltipText: data.settings.dualKpiProperties.topChartToolTipText,
                                top: 0,
                                valueAsPercent: data.topValueAsPercent,
                                width: chartWidth,
                                position: DualKpiChartPositionType["top"],
                                showTextOverlay: data.settings.dualKpiValues.show,
                                showDefaultTextOverlay: !data.settings.dualKpiValues.showKpiValuesTop,
                                defaultTextOverlay: data.settings.dualKpiValues.topChartDefaultKpiValue,
                                fontSizeAutoFormattingTitle: data.settings.dualKpiTitleFormatting.fontSizeAutoFormatting,
                                fontSizeAutoFormattingValue: data.settings.dualKpiValueFormatting.fontSizeAutoFormatting,
                                titleFontSize: data.settings.dualKpiTitleFormatting.fontSize,
                                valueFontSize: data.settings.dualKpiValueFormatting.fontSize,
                                isBoldTitle: data.settings.dualKpiTitleFormatting.isBold,
                                isBoldValue: data.settings.dualKpiValueFormatting.isBold,
                                isItalicTitle: data.settings.dualKpiTitleFormatting.isItalic,
                                isItalicValue: data.settings.dualKpiValueFormatting.isItalic,
                                fontFamilyTitle: data.settings.dualKpiTitleFormatting.fontFamily,
                                fontFamilyValue: data.settings.dualKpiValueFormatting.fontFamily
                            });
                        }
                    }
                    else {
                        this.chartGroupTop.group.classed(DualKpi.INVISIBLE, true);
                        this.chartGroupTop.hoverDataContainer.container.classed(DualKpi.INVISIBLE, true);
                    }
                    if (data.settings.dualKpiProperties.bottomChartShow) {
                        // draw bottom chart
                        this.chartGroupBottom.group.classed(DualKpi.INVISIBLE, false);
                        this.chartGroupBottom.hoverDataContainer.container.classed(DualKpi.INVISIBLE, true);
                        if (data.bottomValues.length > 0) {
                            this.drawChart({
                                element: this.chartGroupBottom,
                                abbreviateValue: data.settings.dualKpiProperties.abbreviateValues,
                                abbreviateHoverValue: data.settings.dualKpiProperties.abbreviateHoverValues,
                                hoverDataPercentType: data.settings.dualKpiProperties.hoverDataPercentType,
                                axisConfig: bottomChartAxisConfig,
                                chartData: data.bottomValues,
                                chartTitle: data.bottomChartName,
                                chartType: data.settings.dualKpiChart.bottomChartType,
                                height: data.settings.dualKpiProperties.bottomChartShow && data.settings.dualKpiProperties.topChartShow ? chartHeight : chartHeight * 2 + chartSpaceBetween,
                                percentChangeStartPoint: bottomChartPercentChangeStartPoint,
                                showZeroLine: data.settings.dualKpiAxis.bottomChartZeroLine,
                                tooltipText: data.settings.dualKpiProperties.bottomChartToolTipText,
                                top: data.settings.dualKpiProperties.bottomChartShow && data.settings.dualKpiProperties.topChartShow ? chartHeight + chartSpaceBetween : 0,
                                valueAsPercent: data.bottomValueAsPercent,
                                width: chartWidth,
                                position: DualKpiChartPositionType["bottom"],
                                showTextOverlay: data.settings.dualKpiValues.show,
                                showDefaultTextOverlay: !data.settings.dualKpiValues.showKpiValuesBottom,
                                defaultTextOverlay: data.settings.dualKpiValues.bottomChartDefaultKpiValue,
                                fontSizeAutoFormattingTitle: data.settings.dualKpiTitleFormatting.fontSizeAutoFormatting,
                                fontSizeAutoFormattingValue: data.settings.dualKpiValueFormatting.fontSizeAutoFormatting,
                                titleFontSize: data.settings.dualKpiTitleFormatting.fontSize,
                                valueFontSize: data.settings.dualKpiValueFormatting.fontSize,
                                isBoldTitle: data.settings.dualKpiTitleFormatting.isBold,
                                isBoldValue: data.settings.dualKpiValueFormatting.isBold,
                                isItalicTitle: data.settings.dualKpiTitleFormatting.isItalic,
                                isItalicValue: data.settings.dualKpiValueFormatting.isItalic,
                                fontFamilyTitle: data.settings.dualKpiTitleFormatting.fontFamily,
                                fontFamilyValue: data.settings.dualKpiValueFormatting.fontFamily
                            });
                        }
                    }
                    else {
                        this.chartGroupBottom.group.classed(DualKpi.INVISIBLE, true);
                        this.chartGroupBottom.hoverDataContainer.container.classed(DualKpi.INVISIBLE, true);
                    }
                    if (wasDataSetRendered) {
                        this.drawBottomContainer(chartWidth, chartHeight, chartTitleSpace, chartSpaceBetween, iconOffset);
                    }
                };
                DualKpi.prototype.displayRootElement = function (isRootElementVisible) {
                    if (isRootElementVisible === void 0) { isRootElementVisible = true; }
                    if (!this.svgRoot) {
                        return;
                    }
                    var display = isRootElementVisible
                        ? null
                        : "none";
                    this.svgRoot.style("display", display);
                };
                DualKpi.prototype.updateViewport = function (viewport) {
                    if (!this.viewport || (this.viewport.width !== viewport.width) || (this.viewport.height !== viewport.height)) {
                        this.svgRoot.attr({
                            width: viewport.width,
                            height: viewport.height
                        });
                    }
                    this.viewport = viewport;
                };
                DualKpi.prototype.enumerateObjectInstances = function (options) {
                    var instances = DualKpiSettings.enumerateObjectInstances(this.data.settings || DualKpiSettings.getDefault(), options).instances;
                    switch (options.objectName) {
                        case "dualKpiColorsBottom": {
                            if (this.data.settings.dualKpiColorsBottom.matchTopChartOptions
                                && instances
                                && instances[0]
                                && instances[0].properties) {
                                delete instances[0].properties["dataColor"];
                                delete instances[0].properties["textColor"];
                                delete instances[0].properties["opacity"];
                            }
                            break;
                        }
                        case "dualKpiTitleFormatting": {
                            if (this.data.settings.dualKpiTitleFormatting.fontSizeAutoFormatting) {
                                delete instances[0].properties["fontSize"];
                            }
                            break;
                        }
                        case "dualKpiValueFormatting": {
                            if (this.data.settings.dualKpiValueFormatting.fontSizeAutoFormatting) {
                                delete instances[0].properties["fontSize"];
                            }
                            break;
                        }
                    }
                    return instances || [];
                };
                DualKpi.validateOpacity = function (opacity) {
                    if (opacity < this.OPACITY_MIN) {
                        return this.OPACITY_MIN;
                    }
                    else if (opacity > this.OPACITY_MAX) {
                        return this.OPACITY_MAX;
                    }
                    else {
                        return opacity;
                    }
                };
                DualKpi.getDaysBetween = function (date1, date2) {
                    var oneDay = 24 * 60 * 60 * 1000; // hours*minutes*seconds*milliseconds
                    var dayRange = Math.round(Math.abs(date1.getTime() - date2.getTime()) / oneDay);
                    return dayRange;
                };
                DualKpi.percentFormatter = function (value, showPlusMinus) {
                    var prefix = value >= 0 ? "+" : "", valueString = (Math.floor(value * 10) / 10) + "%";
                    if (showPlusMinus) {
                        valueString = prefix + valueString;
                    }
                    return valueString;
                };
                DualKpi.getPercentChange = function (startValue, endValue) {
                    if (startValue === 0) {
                        return "n/a";
                    }
                    var diff = endValue - startValue;
                    var percentChange = Math.abs(diff / startValue);
                    if (endValue < startValue) {
                        percentChange = percentChange * -1;
                    }
                    return this.percentFormatter(percentChange * 100, true);
                };
                DualKpi.getPercentChangeStartPoint = function (chartData, percentCalcDate) {
                    if (percentCalcDate !== null) {
                        var closestIndex = 0, percentCalcDateTime = percentCalcDate.getTime(), i = void 0, currTime = void 0;
                        // keep track of closest date to configured date
                        // as soon as we find a date that is more recent than configured date
                        // break and use the last date that was older than configured date.
                        // always break if we find a date that is exactly equal
                        for (i = 0; i < chartData.length; i++) {
                            currTime = chartData[i].date.getTime();
                            if (currTime === percentCalcDateTime) {
                                closestIndex = i;
                                break;
                            }
                            else if (currTime < percentCalcDateTime) {
                                closestIndex = i;
                            }
                            else {
                                break;
                            }
                        }
                        return chartData[closestIndex];
                    }
                    return chartData[0];
                };
                DualKpi.getFormatSymbol = function (format) {
                    var symbolPatterns = [
                        "[$]",
                        "[€]",
                        "[£]",
                        "[¥]",
                        "[₩]",
                        "[%]",
                    ];
                    var symbolMatcher = new RegExp(symbolPatterns.join("|"), "g");
                    //let symbols = [];
                    var match = symbolMatcher.exec(format);
                    if (!match) {
                        return undefined;
                    }
                    else {
                        return match[0];
                    }
                };
                DualKpi.parseSettings = function (dataView) {
                    var settings = DualKpiSettings.parse(dataView);
                    settings.dualKpiColors.opacity = DualKpi.validateOpacity(settings.dualKpiColors.opacity);
                    settings.dualKpiColorsBottom.opacity = DualKpi.validateOpacity(settings.dualKpiColorsBottom.opacity);
                    return settings;
                };
                DualKpi.converter = function (dataView) {
                    var data = {};
                    var topValueFormatSymbol = "";
                    var bottomValueFormatSymbol = "";
                    data.settings = DualKpi.parseSettings(dataView);
                    if (data.settings.dualKpiColorsBottom.matchTopChartOptions) {
                        data.settings.dualKpiColorsBottom.dataColor = data.settings.dualKpiColors.dataColor;
                        data.settings.dualKpiColorsBottom.opacity = data.settings.dualKpiColors.opacity;
                        data.settings.dualKpiColorsBottom.textColor = data.settings.dualKpiColors.textColor;
                    }
                    data.topValues = [];
                    data.bottomValues = [];
                    var axisCol = -1, topValuesCol = -1, bottomValuesCol = -1, warningStateCol = -1, topPercentDateCol = -1, bottomPercentDateCol = -1, rows = [];
                    var metadataColumns = dataView.metadata.columns;
                    for (var i = 0; i < metadataColumns.length; i++) {
                        var col = metadataColumns[i];
                        if (col.roles) {
                            // not else ifs because in a column can have multiple roles
                            if (col.roles["axis"])
                                axisCol = i;
                            if (col.roles["topvalues"]) {
                                topValuesCol = i;
                                data.topChartName = col.displayName;
                                topValueFormatSymbol = this.getFormatSymbol(col.format);
                            }
                            if (col.roles["bottomvalues"]) {
                                bottomValuesCol = i;
                                data.bottomChartName = col.displayName;
                                bottomValueFormatSymbol = this.getFormatSymbol(col.format);
                            }
                            if (col.roles["warningstate"]) {
                                warningStateCol = i;
                            }
                            if (col.roles["toppercentdate"]) {
                                topPercentDateCol = i;
                            }
                            if (col.roles["bottompercentdate"]) {
                                bottomPercentDateCol = i;
                            }
                        }
                    }
                    if (dataView && dataView.table) {
                        rows = dataView.table.rows;
                    }
                    data.topValueAsPercent = topValueFormatSymbol === "%" ? true : false;
                    data.bottomValueAsPercent = bottomValueFormatSymbol === "%" ? true : false;
                    // if percent dates are in data use that, otherwise get from formatting pane/default values
                    // TODO: (defect: 93823) remove unchecked casting and fix. There is not any overload 'new Date()' with 'Date' type argument.
                    data.topPercentCalcDate = topPercentDateCol > -1 && rows[0] ? new Date(rows[0][topPercentDateCol]) : new Date(data.settings.dualKpiProperties.topPercentCalcDate);
                    data.bottomPercentCalcDate = bottomPercentDateCol > -1 && rows[0] ? new Date(rows[0][bottomPercentDateCol]) : new Date(data.settings.dualKpiProperties.bottomPercentCalcDate);
                    for (var i = 0; i < rows.length; i++) {
                        var date = null;
                        if (axisCol > -1) {
                            var timestamp = Date.parse(rows[i][axisCol]);
                            if (!isNaN(timestamp)) {
                                date = new Date(timestamp);
                            }
                            else {
                                continue;
                            }
                        }
                        else {
                            date = new Date();
                        }
                        var topValue = topValuesCol > -1 ? rows[i][topValuesCol] : 0;
                        var bottomValue = bottomValuesCol > -1 ? rows[i][bottomValuesCol] : 0;
                        if (data.topValueAsPercent) {
                            topValue *= 100;
                        }
                        if (data.bottomValueAsPercent) {
                            bottomValue *= 100;
                        }
                        data.topValues.push({
                            date: date,
                            value: topValue
                        });
                        data.bottomValues.push({
                            date: date,
                            value: bottomValue
                        });
                    }
                    if ((warningStateCol > -1) && (rows.length > 0)) {
                        data.warningState = rows[rows.length - 1][warningStateCol];
                    }
                    data.topValues = _.sortBy(data.topValues, "date");
                    data.bottomValues = _.sortBy(data.bottomValues, "date");
                    return data;
                };
                DualKpi.prototype.createHoverDataContainer = function (chartGroup) {
                    var hoverDataContainer = chartGroup.append("g")
                        .attr("class", "hover-data-container")
                        .classed(DualKpi.INVISIBLE, true);
                    var date = hoverDataContainer.append("text")
                        .attr("class", "hover-text date")
                        .text("0");
                    var text = hoverDataContainer.append("text")
                        .attr("class", "hover-text value")
                        .attr("text-anchor", "middle")
                        .text("0");
                    var percent = hoverDataContainer.append("text")
                        .attr("class", "hover-text percent")
                        .attr("text-anchor", "end")
                        .text("0");
                    return {
                        container: hoverDataContainer,
                        date: date,
                        text: text,
                        percent: percent
                    };
                };
                DualKpi.prototype.updateHoverDataContainer = function (hoverDataContainer, chartBottom, chartLeft, chartWidth, isTopChart) {
                    var textColor = isTopChart ? this.data.settings.dualKpiColors.textColor : this.data.settings.dualKpiColorsBottom.textColor;
                    var hoverDate = hoverDataContainer.date;
                    var centerX = chartWidth / 2;
                    if (chartWidth < 300) {
                        centerX *= 0.85;
                    }
                    hoverDate
                        .attr("class", "hover-text date")
                        .classed(this.sizeCssClass, true)
                        .attr("fill", textColor)
                        .text("0");
                    var hoverValue = hoverDataContainer.text;
                    hoverValue
                        .attr("class", "hover-text value")
                        .classed(this.sizeCssClass, true)
                        .attr("transform", "translate(" + centerX + ",0)")
                        .attr("fill", textColor)
                        .text("0");
                    var hoverPercent = hoverDataContainer.percent;
                    hoverPercent
                        .attr("class", "hover-text percent")
                        .classed(this.sizeCssClass, true)
                        .text("0")
                        .attr("fill", textColor)
                        .attr("transform", "translate(" + (chartWidth) + ",0)");
                    hoverDataContainer.container
                        .attr("transform", "translate(" + 0 + "," + (chartBottom + this.titleHeight - 2) + ")");
                };
                DualKpi.prototype.showHoverData = function (hoverDataContainer, dataPoint, latestValue, hoverDataPercentType, valueAsPercent, abbreviateHoverValue) {
                    var _this = this;
                    var hoverDate = hoverDataContainer.date;
                    hoverDate
                        .datum(dataPoint)
                        .text(function (d) { return _this.timeFormatter(d.date); });
                    var hoverValue = hoverDataContainer.text;
                    hoverValue
                        .datum(dataPoint)
                        .text(function (d) {
                        var value = abbreviateHoverValue ? _this.valueFormatter(d.value) : _this.commaNumberFormatter(Math.round(d.value));
                        if (valueAsPercent) {
                            return DualKpi.percentFormatter(d.value);
                        }
                        return value;
                    });
                    var hoverPercent = hoverDataContainer.percent;
                    hoverPercent
                        .datum(dataPoint)
                        .text(function (d) {
                        if (valueAsPercent) {
                            var value = hoverDataPercentType === PercentType.lastDate ? latestValue - d.value
                                : d.value - latestValue;
                            return DualKpi.percentFormatter(value);
                        }
                        var leftValue = hoverDataPercentType === PercentType.lastDate ? d.value : latestValue, rightValue = hoverDataPercentType === PercentType.lastDate ? latestValue : d.value;
                        return DualKpi.getPercentChange(leftValue, rightValue);
                    });
                    this.bottomContainer.bottomContainer.classed("hidden", true);
                    hoverDataContainer.container.classed(DualKpi.INVISIBLE, false);
                };
                DualKpi.prototype.hideHoverData = function (hoverDataContainer, hoverLine) {
                    hoverDataContainer.container.classed(DualKpi.INVISIBLE, true);
                    this.bottomContainer.bottomContainer.classed("hidden", false);
                    if (hoverLine) {
                        hoverLine.classed(DualKpi.INVISIBLE, true);
                    }
                };
                /*
                *   to show tooltip information on mobile we show a popup on touch event
                */
                DualKpi.prototype.showMobileTooltip = function (message) {
                    var _this = this;
                    if (!this.mobileTooltip) {
                        this.mobileTooltip = d3.select(this.target[0]).append("div")
                            .classed({ "hidden": true, "mobile-tooltip": true });
                        this.svgRoot.on("touchstart", function () {
                            _this.hideMobileTooltip();
                        });
                        this.mobileTooltip.on("touchstart", function () {
                            _this.hideMobileTooltip();
                        });
                        this.touchEventsEnabled = true;
                    }
                    // prevent hide from being called, and prevent hover interaction from occuring on same event
                    d3.event.stopPropagation();
                    this.mobileTooltip.text(message);
                    this.mobileTooltip.classed("hidden", false);
                };
                DualKpi.prototype.hideMobileTooltip = function () {
                    this.mobileTooltip.classed("hidden", true);
                };
                DualKpi.prototype.drawBottomContainer = function (chartWidth, chartHeight, chartTitleSpace, chartSpaceBetween, iconOffset) {
                    //let warningIconShowing = false;
                    var infoIconShowing = false;
                    var chartTitleElement = this.bottomContainer.chartTitleElement
                        .attr("class", "title")
                        .classed(this.sizeCssClass, true)
                        .text(this.data.settings.dualKpiProperties.titleText);
                    var iconWidth = 22;
                    var iconScaleTransform = "";
                    var iconY = (-chartTitleSpace + (chartTitleSpace / 2) + iconOffset);
                    if (this.size === DualKpiSize.small || this.size === DualKpiSize.extrasmall || this.size === DualKpiSize.supersmall) {
                        iconScaleTransform = "scale(0.75)";
                        iconWidth = 16;
                    }
                    // add warning icon
                    if (this.data.warningState < 0) {
                        this.createWarningMessage(chartTitleElement, iconY, iconScaleTransform, iconWidth);
                    }
                    // add info icon
                    if (this.data.topValues.length > 0) {
                        var today = new Date();
                        var dataDaysOld = DualKpi.getDaysBetween(this.data.topValues[this.data.topValues.length - 1].date, today);
                        if (dataDaysOld >= this.data.settings.dualKpiProperties.staleDataThreshold && this.data.settings.dualKpiProperties.showStaleDataWarning) {
                            infoIconShowing = true;
                            this.createInfoMessage(iconY, iconScaleTransform, iconWidth, chartWidth, dataDaysOld);
                        }
                        else {
                            this.hideInfoMessage();
                        }
                        // add day range text
                        var dayRange = DualKpi.getDaysBetween(this.data.topValues[0].date, this.data.topValues[this.data.topValues.length - 1].date);
                        var dayRangeElement = this.bottomContainer.dateRangeText;
                        dayRangeElement
                            .attr("class", "date-range-text")
                            .classed(this.sizeCssClass, true)
                            .text("last " + dayRange + " days");
                        var dayRangeLeft = chartWidth - 8;
                        if (infoIconShowing) {
                            dayRangeLeft -= (iconWidth); // width of icon + 8px padding
                        }
                        dayRangeElement.attr("transform", "translate(" + (dayRangeLeft) + ",0)");
                    }
                    this.bottomContainer.bottomContainer.attr("transform", "translate(5," + (this.viewport.height - 5) + ")");
                    this.bottomContainer.bottomContainer.classed(DualKpi.INVISIBLE, false);
                };
                DualKpi.prototype.createWarningMessage = function (chartTitleElement, iconY, iconScaleTransform, iconWidth) {
                    var _this = this;
                    var warning = this.bottomContainer.warning;
                    warning.group
                        .attr("transform", "translate(0," + (iconY) + ")");
                    var warningIcon = warning.icon;
                    warningIcon
                        .attr({
                        "d": "M24,24H8l8-16L24,24z M9.7,23h12.6L16,10.4L9.7,23z M16.5,19.8h-1v-5.4h1V19.8z M16.5,20.8v1.1h-1v-1.1H16.5z",
                        "fill": "#E81123",
                        "stroke": "transparent",
                        "stroke-width": "5",
                        "class": "warning-icon",
                        "transform": iconScaleTransform
                    })
                        .classed(this.sizeCssClass, true);
                    var warningTitle = warning.title;
                    warningTitle
                        .text(this.data.settings.dualKpiProperties.warningTooltipText);
                    // move title over to account for icon
                    chartTitleElement.attr("transform", "translate(" + (iconWidth + 6) + ",0)");
                    warning.group.on("touchstart", function () { return _this.showMobileTooltip(_this.data.settings.dualKpiProperties.warningTooltipText); });
                };
                DualKpi.prototype.createInfoMessage = function (iconY, iconScaleTransform, iconWidth, chartWidth, dataDaysOld) {
                    var _this = this;
                    var infoMessage = "Data is " + dataDaysOld + " days old. " + this.data.settings.dualKpiProperties.staleDataTooltipText;
                    var info = this.bottomContainer.info;
                    info.group
                        .attr("transform", "translate(" + (chartWidth - iconWidth - 8) + "," + (iconY) + ")");
                    var infoIcon = info.icon;
                    infoIcon
                        .attr({
                        "d": "M24,16c0,1.4-0.4,2.8-1,4c-0.7,1.2-1.7,2.2-2.9,2.9c-1.2,0.7-2.5,1-4,1s-2.8-0.4-4-1c-1.2-0.7-2.2-1.7-2.9-2.9 C8.4,18.8,8,17.4,8,16c0-1.5,0.4-2.8,1.1-4c0.8-1.2,1.7-2.2,2.9-2.9S14.6,8,16,8s2.8,0.3,4,1.1c1.2,0.7,2.2,1.7,2.9,2.9 C23.6,13.2,24,14.5,24,16z M12.6,22c1.1,0.6,2.2 0.9,3.4,0.9s2.4-0.3,3.5-0.9c1-0.6,1.9-1.5,2.5-2.6c0.6-1,1-2.2,1-3.4 s-0.3-2.4-1-3.5s-1.5-1.9-2.5-2.5c-1.1-0.6-2.2-1-3.5-1s-2.4,0.4-3.4,1c-1.1,0.6-1.9,1.4-2.6,2.5c-0.6,1.1-0.9,2.2-0.9,3.5 c0,1.2,0.3,2.4,0.9,3.4C10.6,20.5,11.4,21.4,12.6,22z M16.5,17.6h-1v-5.4h1V17.6z M16.5 19.7h-1v-1.1h1V19.7z",
                        "fill": "#3599B8",
                        "stroke": "transparent",
                        "stroke-width": "5",
                        "class": "info-icon",
                        "transform": iconScaleTransform
                    })
                        .classed(this.sizeCssClass, true)
                        .classed("hidden", false);
                    var infoTitle = info.title;
                    infoTitle
                        .text(infoMessage);
                    info.group.on("touchstart", function () { return _this.showMobileTooltip(infoMessage); });
                };
                DualKpi.prototype.hideInfoMessage = function () {
                    var info = this.bottomContainer.info;
                    info.icon.classed("hidden", true);
                };
                DualKpi.prototype.drawChart = function (options) {
                    var _this = this;
                    var chartData = options.chartData;
                    var axisConfig = options.axisConfig;
                    var latestValue = chartData[chartData.length - 1].value, isTopChart = options.position === DualKpiChartPositionType.top, dataColor = isTopChart ? this.data.settings.dualKpiColors.dataColor : this.data.settings.dualKpiColorsBottom.dataColor, chartOpacity = isTopChart ? this.data.settings.dualKpiColors.opacity : this.data.settings.dualKpiColorsBottom.opacity;
                    var target = this.target[0];
                    var targetPadding = parseInt(this.target.css("padding-left"), 10) || 0;
                    var margin = {
                        top: 7,
                        right: 0,
                        bottom: 0,
                        left: this.chartLeftMargin
                    };
                    if (this.size === DualKpiSize.medium || this.size === DualKpiSize.large) {
                        margin.left = 40;
                    }
                    var calcWidth = options.width - margin.right - margin.left, calcHeight = options.height - margin.top - margin.bottom, minValue = d3.min(chartData, function (d) { return d.value; }) || 0, maxValue = d3.max(chartData, function (d) { return d.value; }) || 0;
                    var axisMinValue = axisConfig.min !== null && axisConfig.min !== undefined ? axisConfig.min : minValue;
                    var axisMaxValue = axisConfig.max !== null && axisConfig.max !== undefined ? axisConfig.max : maxValue;
                    var xScale = d3.time.scale()
                        .domain(d3.extent(chartData, function (d) { return d.date; }))
                        .range([0, calcWidth]);
                    var yScale = d3.scale.linear()
                        .domain([axisMinValue, axisMaxValue])
                        .clamp(true)
                        .range([calcHeight, 0]);
                    var yAxis = d3.svg.axis()
                        .scale(yScale)
                        .tickValues([axisMinValue, axisMaxValue])
                        .tickFormat(function (d) {
                        var axisTickLabel = String(_this.axisNumberFormatter(d));
                        if (options.valueAsPercent) {
                            axisTickLabel = axisTickLabel + "%";
                        }
                        return axisTickLabel;
                    })
                        .orient("left");
                    var seriesRenderer, fill, stroke, strokeWidth;
                    if (options.chartType === "area") {
                        seriesRenderer = d3.svg.area()
                            .x(function (d) { return xScale(d.date || new Date()); })
                            .y0(calcHeight)
                            .y1(function (d) { return yScale(d.value || 0); });
                        fill = dataColor;
                        stroke = "none";
                        strokeWidth = 0;
                    }
                    else {
                        seriesRenderer = d3.svg.line()
                            .x(function (d) { return xScale(d.date || new Date()); })
                            .y(function (d) { return yScale(d.value || 0); });
                        fill = "none";
                        stroke = dataColor;
                        strokeWidth = 2;
                    }
                    var chartGroup = options.element;
                    chartGroup.group
                        .attr("transform", "translate(" + margin.left + "," + (options.top + margin.top) + ")");
                    var chartArea = chartGroup.area;
                    chartArea
                        .datum(chartData)
                        .attr({
                        "style": "opacity: " + (chartOpacity / 100),
                        "fill": fill,
                        "stroke": stroke,
                        "stroke-width": strokeWidth,
                        "d": seriesRenderer
                    });
                    var zeroAxis = chartGroup.zeroAxis;
                    var zeroPointOnAxis = axisMinValue <= 0 && axisMaxValue >= 0 ? true : false;
                    // DRAW line for x axis at zero position
                    // if formatting option for zero line set to true
                    // and if a value of zero is on the y-axis
                    if (options.showZeroLine && zeroPointOnAxis) {
                        var axisLine = d3.svg.line()
                            .x(function (d) { return xScale(d.date); })
                            .y(function (d) { return yScale(0); });
                        zeroAxis
                            .datum(chartData)
                            .classed("hidden", false)
                            .attr({
                            "d": axisLine
                        });
                    }
                    else {
                        zeroAxis
                            .classed("hidden", true);
                    }
                    var axis = chartGroup.yAxis;
                    axis
                        .attr("class", "axis")
                        .classed(this.sizeCssClass, true)
                        .call(yAxis);
                    /* Add elements for hover behavior ******************************************************/
                    var hoverLine = chartGroup.hoverLine;
                    hoverLine
                        .classed(DualKpi.INVISIBLE, true)
                        .attr({
                        "x1": 0,
                        "x2": 0,
                        "y1": 0,
                        "y2": calcHeight,
                        "stroke-width": 1,
                        "stroke": "#777"
                    });
                    var chartBottom = margin.top + calcHeight;
                    var chartLeft = margin.left;
                    var hoverDataContainer = options.element.hoverDataContainer;
                    this.updateHoverDataContainer(hoverDataContainer, chartBottom, chartLeft, calcWidth, isTopChart);
                    this.dispatch.on("onDualKpiMouseMove." + options.position, function (_a) {
                        var leftPosition = _a[0], topPosition = _a[1];
                        var areaScale = DualKpi.getScale(target);
                        var maxWidth = options.width - margin.left;
                        leftPosition = leftPosition / areaScale.x - margin.left - targetPadding;
                        var x = xScale.invert(leftPosition);
                        var i = _this.dataBisector(chartData, x, 1);
                        var dataPoint = chartData[i];
                        if ((leftPosition > 0) &&
                            (topPosition > 0) &&
                            (leftPosition < maxWidth) &&
                            (topPosition < (options.height * 2 + 15)) &&
                            dataPoint) {
                            hoverLine.attr("transform", "translate(" + leftPosition + ", 0)");
                            hoverLine.classed(DualKpi.INVISIBLE, false);
                            var value = options.hoverDataPercentType === PercentType.lastDate ? chartData[chartData.length - 1].value
                                : options.hoverDataPercentType === PercentType.firstDate ? chartData[0].value
                                    : chartData[i - 1] ? chartData[i - 1].value : 0;
                            _this.showHoverData(hoverDataContainer, dataPoint, value, options.hoverDataPercentType, options.valueAsPercent, options.abbreviateHoverValue);
                        }
                        else {
                            _this.hideHoverData(hoverDataContainer, hoverLine);
                        }
                    });
                    this.dispatch.on("onDualKpiMouseOut." + options.position, function () {
                        _this.hideHoverData(hoverDataContainer, hoverLine);
                    });
                    if (options.showTextOverlay) {
                        this.addOverlayText(options, latestValue, calcHeight, calcWidth, isTopChart);
                    }
                    else {
                        options.element.chartOverlay.rect.classed(DualKpi.INVISIBLE, true);
                        options.element.chartOverlay.title.classed(DualKpi.INVISIBLE, true);
                        options.element.chartOverlay.text.classed(DualKpi.INVISIBLE, true);
                    }
                };
                DualKpi.getScale = function (element) {
                    var clientRect = element.getBoundingClientRect();
                    return {
                        x: clientRect.width / element.offsetWidth,
                        y: clientRect.height / element.offsetHeight
                    };
                };
                DualKpi.prototype.applyFontSize = function (element, fontSizeAutoFormattingTitle, fontSize) {
                    if (!fontSizeAutoFormattingTitle) {
                        element.attr("font-size", fontSize);
                    }
                    else {
                        element.classed(this.sizeCssClass, true);
                    }
                };
                DualKpi.prototype.applyBold = function (element, isBold) {
                    if (isBold) {
                        element.attr("font-weight", "bold");
                    }
                    else {
                        element.attr("font-weight", null);
                    }
                };
                DualKpi.prototype.applyItalic = function (element, isItalic) {
                    if (isItalic) {
                        element.attr("font-style", "italic");
                    }
                    else {
                        element.attr("font-style", null);
                    }
                };
                DualKpi.prototype.applyFontFamily = function (element, fontFamily) {
                    element.style("font-family", fontFamily);
                };
                DualKpi.prototype.applyTextStyle = function (element, options, isTitle) {
                    var fontSizeAutoFormatting = isTitle ? options.fontSizeAutoFormattingTitle : options.fontSizeAutoFormattingValue, fontSize = isTitle ? options.titleFontSize : options.valueFontSize, isBold = isTitle ? options.isBoldTitle : options.isBoldValue, isItalic = isTitle ? options.isItalicTitle : options.isItalicValue, fontFamily = isTitle ? options.fontFamilyTitle : options.fontFamilyValue;
                    this.applyFontSize(element, fontSizeAutoFormatting, fontSize);
                    this.applyBold(element, isBold);
                    this.applyItalic(element, isItalic);
                    this.applyFontFamily(element, fontFamily);
                };
                DualKpi.prototype.addOverlayText = function (options, latestValue, calcHeight, calcWidth, isTopChart) {
                    var _this = this;
                    var textColor = isTopChart ? this.data.settings.dualKpiColors.textColor : this.data.settings.dualKpiColorsBottom.textColor;
                    var chartData = options.chartData;
                    var chartGroup = options.element;
                    var percentChange = DualKpi.getPercentChange(options.percentChangeStartPoint.value, chartData[chartData.length - 1].value);
                    var format;
                    for (var col = 0; col < this.dataView.table.columns.length; col++) {
                        var column = this.dataView.table.columns[col];
                        if (column.roles["topvalues"] && isTopChart) {
                            format = column.format;
                            break;
                        }
                        if (column.roles["bottomvalues"] && !isTopChart) {
                            format = column.format;
                            break;
                        }
                    }
                    var formatter = valueFormatter.create({
                        format: format,
                        precision: this.data.settings.dualKpiValueFormatting.precision,
                        value: this.data.settings.dualKpiValueFormatting.displayUnits || latestValue,
                        displayUnitSystemType: powerbi.DisplayUnitSystemType.WholeUnits,
                    });
                    var formattedValue;
                    if (options.showDefaultTextOverlay) {
                        formattedValue = options.defaultTextOverlay;
                    }
                    else {
                        formattedValue = options.abbreviateValue ? this.valueFormatter(latestValue) : formatter.format(latestValue);
                    }
                    if (options.valueAsPercent) {
                        formattedValue = DualKpi.percentFormatter(latestValue);
                        // if value is a percent, only show difference changed, not percent of percent
                        percentChange = DualKpi.percentFormatter(chartData[chartData.length - 1].value - options.percentChangeStartPoint.value, true);
                    }
                    var chartOverlay = chartGroup.chartOverlay;
                    var dataTitle = chartOverlay.title;
                    dataTitle
                        .classed(DualKpi.INVISIBLE, true)
                        .attr("class", "data-title")
                        .attr("fill", textColor)
                        .text(options.showDefaultTextOverlay ? "" : options.chartTitle + " (" + percentChange + ")");
                    this.applyTextStyle(dataTitle, options, true);
                    var dataValue = chartOverlay.text;
                    dataValue
                        .classed(DualKpi.INVISIBLE, true)
                        .attr("class", "data-value")
                        .attr("fill", textColor)
                        .text(formattedValue);
                    this.applyTextStyle(dataValue, options);
                    var dataTitleHorzCentering = calcWidth / 2;
                    var dataValueHorzCentering = calcWidth / 2;
                    var verticalMargin = DualKpi.DefaultValueSizes[this.sizeCssClass];
                    // apply centerings, then unhide text
                    dataTitle.attr("transform", "translate(" + dataTitleHorzCentering + ", 0)");
                    dataValue.attr("transform", "translate(" + dataValueHorzCentering + ", " + verticalMargin + ")");
                    var verticalCentering = (calcHeight / 2) - verticalMargin / 2; // bump slightly above perfectly vertically centered on chart
                    var horizontalCentering = 0;
                    chartOverlay.group
                        .attr("transform", "translate(" + horizontalCentering + ", " + verticalCentering + ")");
                    dataTitle.classed(DualKpi.INVISIBLE, false);
                    dataValue.classed(DualKpi.INVISIBLE, false);
                    // set rect dimensions
                    // add rect to overlay section so that tooltip shows up more easily
                    var overlayRect = chartOverlay.rect;
                    // add tooltip
                    var percentChangeDesc = percentChange;
                    if (!this.data.settings.dualKpiProperties.shortKpiTooltip) {
                        percentChangeDesc += " change since " + this.timeFormatter(options.percentChangeStartPoint.date);
                    }
                    var overlayTooltipText;
                    if (options.showDefaultTextOverlay) {
                        overlayTooltipText = formattedValue;
                    }
                    else {
                        overlayTooltipText = options.tooltipText + " " + percentChangeDesc;
                    }
                    var overlayTooltip = chartOverlay.rectTitle;
                    overlayTooltip
                        .text(overlayTooltipText);
                    var dataTitleProps = textMeasurementService.getSvgMeasurementProperties(dataTitle.node());
                    var dataValueProps = textMeasurementService.getSvgMeasurementProperties(dataValue.node());
                    var dataTitleWidth = textMeasurementService.measureSvgTextWidth(dataTitleProps);
                    var dataTitleHeight = textMeasurementService.measureSvgTextHeight(dataTitleProps);
                    var dataValueWidth = textMeasurementService.measureSvgTextWidth(dataValueProps);
                    var dataValueHeight = textMeasurementService.measureSvgTextHeight(dataValueProps);
                    var dataWidth = Math.max(dataTitleWidth, dataValueWidth);
                    var dataHeight = dataTitleHeight + dataValueHeight + (verticalMargin);
                    overlayRect
                        .attr("width", dataWidth)
                        .attr("height", dataHeight)
                        .attr("transform", "translate(" + (dataTitleHorzCentering - (dataWidth / 2)) + "," + (-dataValueHeight) + ")");
                    overlayRect.on("touchstart", function () { return _this.showMobileTooltip(overlayTooltipText); });
                    overlayRect.on("mousemove", function () {
                        if (_this.touchEventsEnabled) {
                            d3.event.stopPropagation();
                        }
                    });
                };
                DualKpi.capabilities = {
                    dataRoles: [
                        {
                            displayName: "Axis",
                            name: "axis",
                            kind: powerbi.VisualDataRoleKind.Grouping
                        },
                        {
                            displayName: "Top values",
                            name: "topvalues",
                            kind: powerbi.VisualDataRoleKind.Measure,
                            requiredTypes: [{
                                    "numeric": true
                                }, {
                                    "integer": true
                                }]
                        },
                        {
                            displayName: "Bottom values",
                            name: "bottomvalues",
                            kind: powerbi.VisualDataRoleKind.Measure,
                            requiredTypes: [{
                                    "numeric": true
                                }, {
                                    "integer": true
                                }]
                        },
                        {
                            displayName: "Warning state",
                            name: "warningstate",
                            kind: powerbi.VisualDataRoleKind.Measure
                        },
                        {
                            displayName: "Top - % change start date",
                            name: "toppercentdate",
                            kind: powerbi.VisualDataRoleKind.Grouping
                        },
                        {
                            displayName: "Bottom - % change start date",
                            name: "bottompercentdate",
                            kind: powerbi.VisualDataRoleKind.Grouping
                        },
                    ],
                    dataViewMappings: [{
                            conditions: [
                                {
                                    "axis": { max: 1 },
                                    "topvalues": { max: 1 },
                                    "bottomvalues": { max: 1 },
                                    "warningstate": { max: 1 },
                                    "toppercentdate": { max: 1 },
                                    "bottompercentdate": { max: 1 }
                                }
                            ],
                            table: {
                                rows: {
                                    for: {
                                        in: "axis"
                                    },
                                    dataReductionAlgorithm: { top: { count: 1000 } }
                                }
                            },
                            categorical: {
                                categories: {
                                    select: [
                                        { for: { in: "axis" } },
                                    ],
                                    dataReductionAlgorithm: { top: { count: 1000 } }
                                },
                                values: {
                                    select: [{
                                            bind: {
                                                to: "axis"
                                            }
                                        }, {
                                            bind: {
                                                to: "topvalues"
                                            }
                                        }, {
                                            bind: {
                                                to: "bottomvalues"
                                            }
                                        }, {
                                            bind: {
                                                to: "warningstate"
                                            }
                                        }, {
                                            bind: {
                                                to: "toppercentdate"
                                            }
                                        }, {
                                            bind: {
                                                to: "bottompercentdate"
                                            }
                                        }]
                                }
                            },
                        }],
                    objects: {
                        "dualKpiProperties": {
                            "displayName": "Dual KPI Properties",
                            "properties": {
                                "topChartShow": {
                                    "displayName": "Show top chart",
                                    "type": {
                                        "bool": true
                                    }
                                },
                                "bottomChartShow": {
                                    "displayName": "Show bottom chart",
                                    "type": {
                                        "bool": true
                                    }
                                },
                                "titleText": {
                                    "displayName": "Title text",
                                    "type": {
                                        "text": true
                                    }
                                },
                                "abbreviateValues": {
                                    "displayName": "Abbreviate values",
                                    "type": {
                                        "bool": true
                                    }
                                },
                                "abbreviateHoverValues": {
                                    "displayName": "Abbreviate hover values only",
                                    "type": {
                                        "bool": true
                                    }
                                },
                                "shortKpiTooltip": {
                                    "displayName": "Short KPI tooltip",
                                    "type": {
                                        "bool": true
                                    }
                                },
                                "hoverDataPercentType": {
                                    "displayName": "Variance type",
                                    "type": { "enumeration": DualKpiHoverDataPercentType.type }
                                },
                                "topChartToolTipText": {
                                    "displayName": "Top chart tooltip text",
                                    "type": {
                                        "text": true
                                    }
                                },
                                "bottomChartToolTipText": {
                                    "displayName": "Bottom chart tooltip text",
                                    "type": {
                                        "text": true
                                    }
                                },
                                "warningTooltipText": {
                                    "displayName": "Warning tooltip text",
                                    "type": {
                                        "text": true
                                    }
                                },
                                "showStaleDataWarning": {
                                    "displayName": "Show stale data warning",
                                    "type": {
                                        "bool": true
                                    }
                                },
                                "staleDataTooltipText": {
                                    "displayName": "Stale data tooltip text",
                                    "type": {
                                        "text": true
                                    }
                                },
                                "staleDataThreshold": {
                                    "displayName": "Stale data threshold",
                                    "type": {
                                        "numeric": true
                                    }
                                },
                                "topPercentCalcDate": {
                                    "displayName": "Top - % change start date (mm/dd/yyyy)",
                                    "type": {
                                        "text": true
                                    }
                                },
                                "bottomPercentCalcDate": {
                                    "displayName": "Bottom - % change start date (mm/dd/yyyy)",
                                    "type": {
                                        "text": true
                                    }
                                }
                            }
                        },
                        "dualKpiValues": {
                            "displayName": "Dual KPI Values",
                            "properties": {
                                "show": {
                                    "type": {
                                        "bool": true
                                    }
                                },
                                "showKpiValuesTop": {
                                    "displayName": "Show top chart KPI values",
                                    "type": {
                                        "bool": true
                                    }
                                },
                                "topChartDefaultKpiValue": {
                                    "displayName": "Top chart default KPI text",
                                    "type": {
                                        "text": true
                                    }
                                },
                                "showKpiValuesBottom": {
                                    "displayName": "Show bottom chart KPI values",
                                    "type": {
                                        "bool": true
                                    }
                                },
                                "bottomChartDefaultKpiValue": {
                                    "displayName": "Bottom chart default KPI text",
                                    "type": {
                                        "text": true
                                    }
                                }
                            }
                        },
                        "dualKpiTitleFormatting": {
                            "displayName": "Dual KPI Title formatting",
                            "properties": {
                                "fontSizeAutoFormatting": {
                                    "displayName": "Auto text size",
                                    "type": {
                                        "bool": true
                                    }
                                },
                                "fontSize": {
                                    "displayName": "Value text size",
                                    "type": {
                                        "formatting": {
                                            "fontSize": true
                                        }
                                    }
                                },
                                "isBold": {
                                    "displayName": "Bold",
                                    "type": {
                                        "bool": true
                                    }
                                },
                                "isItalic": {
                                    "displayName": "Italic",
                                    "type": {
                                        "bool": true
                                    }
                                },
                                "fontFamily": {
                                    "displayName": "Font family",
                                    "type": {
                                        "formatting": {
                                            "fontFamily": true
                                        }
                                    }
                                }
                            }
                        },
                        "dualKpiValueFormatting": {
                            "displayName": "Dual KPI Value formatting",
                            "properties": {
                                "fontSizeAutoFormatting": {
                                    "displayName": "Auto text size",
                                    "type": {
                                        "bool": true
                                    }
                                },
                                "fontSize": {
                                    "displayName": "Value text size",
                                    "type": {
                                        "formatting": {
                                            "fontSize": true
                                        }
                                    }
                                },
                                "isBold": {
                                    "displayName": "Bold",
                                    "type": {
                                        "bool": true
                                    }
                                },
                                "isItalic": {
                                    "displayName": "Italic",
                                    "type": {
                                        "bool": true
                                    }
                                },
                                "fontFamily": {
                                    "displayName": "Font family",
                                    "type": {
                                        "formatting": {
                                            "fontFamily": true
                                        }
                                    }
                                },
                                "displayUnits": {
                                    "displayName": "Display units",
                                    "type": {
                                        "formatting": {
                                            "labelDisplayUnits": true
                                        }
                                    }
                                },
                                "precision": {
                                    "displayName": "Decimal Places",
                                    "description": "Select the number of decimal places to display",
                                    "type": {
                                        "numeric": true
                                    },
                                    "placeHolderText": "Auto",
                                }
                            }
                        },
                        "dualKpiColors": {
                            "displayName": "Dual KPI Top Chart Colors",
                            "properties": {
                                "dataColor": {
                                    "displayName": "Data color",
                                    "type": {
                                        "fill": {
                                            "solid": {
                                                "color": true
                                            }
                                        }
                                    }
                                },
                                "textColor": {
                                    "displayName": "Text color",
                                    "type": {
                                        "fill": {
                                            "solid": {
                                                "color": true
                                            }
                                        }
                                    }
                                },
                                "opacity": {
                                    "displayName": "Chart opacity",
                                    "type": {
                                        "numeric": true
                                    }
                                }
                            }
                        },
                        "dualKpiColorsBottom": {
                            "displayName": "Dual KPI Bottom Chart Colors",
                            "properties": {
                                "matchTopChartOptions": {
                                    "displayName": "Match top chart settings",
                                    "type": {
                                        "bool": true
                                    }
                                },
                                "dataColor": {
                                    "displayName": "Data color",
                                    "type": {
                                        "fill": {
                                            "solid": {
                                                "color": true
                                            }
                                        }
                                    }
                                },
                                "textColor": {
                                    "displayName": "Text color",
                                    "type": {
                                        "fill": {
                                            "solid": {
                                                "color": true
                                            }
                                        }
                                    }
                                },
                                "opacity": {
                                    "displayName": "Chart opacity",
                                    "type": {
                                        "numeric": true
                                    }
                                }
                            }
                        },
                        "dualKpiAxis": {
                            "displayName": "Dual KPI Axis Settings",
                            "properties": {
                                "topChartAxisMin": {
                                    "displayName": "Top chart axis min",
                                    "type": {
                                        "numeric": true
                                    }
                                },
                                "topChartAxisMax": {
                                    "displayName": "Top chart axis max",
                                    "type": {
                                        "numeric": true
                                    }
                                },
                                "bottomChartAxisMin": {
                                    "displayName": "Bottom chart axis min",
                                    "type": {
                                        "numeric": true
                                    }
                                },
                                "bottomChartAxisMax": {
                                    "displayName": "Bottom chart axis max",
                                    "type": {
                                        "numeric": true
                                    }
                                },
                                "topChartZeroLine": {
                                    "displayName": "Top chart zero line",
                                    "type": {
                                        "bool": true
                                    }
                                },
                                "bottomChartZeroLine": {
                                    "displayName": "Bottom chart zero line",
                                    "type": {
                                        "bool": true
                                    }
                                }
                            }
                        },
                        "dualKpiChart": {
                            "displayName": "Dual KPI Chart Type",
                            "properties": {
                                "topChartType": {
                                    "displayName": "Top chart type",
                                    "type": { "enumeration": DualKpiChartType.type }
                                },
                                "bottomChartType": {
                                    "displayName": "Bottom chart type",
                                    "type": { "enumeration": DualKpiChartType.type }
                                }
                            }
                        }
                    }
                };
                DualKpi.DefaultTitleSizes = {
                    "super-small": 10,
                    "extra-small": 11,
                    "small": 12,
                    "medium": 14,
                    "large": 16
                };
                DualKpi.DefaultValueSizes = {
                    "super-small": 14,
                    "extra-small": 18,
                    "small": 26,
                    "medium": 32,
                    "large": 40
                };
                DualKpi.INVISIBLE = "invisible";
                DualKpi.OPACITY_MIN = 0;
                DualKpi.OPACITY_MAX = 100;
                return DualKpi;
            }()); /*close IVisual*/
            PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.DualKpi = DualKpi;
        })(PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 || (visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = {}));
    })(visuals = powerbi.visuals || (powerbi.visuals = {}));
})(powerbi || (powerbi = {})); /*close export*/
var powerbi;
(function (powerbi) {
    var visuals;
    (function (visuals) {
        var plugins;
        (function (plugins) {
            plugins.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0 = {
                name: 'PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0',
                class: 'PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0',
                capabilities: powerbi.visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.DualKpi.capabilities,
                custom: true,
                create: function () { return new powerbi.visuals.PBI_CV_3C80B1F2_09AF_4123_8E99_C3CBC46B23E0.DualKpi(); }
            };
        })(plugins = visuals.plugins || (visuals.plugins = {}));
    })(visuals = powerbi.visuals || (powerbi.visuals = {}));
})(powerbi || (powerbi = {}));
